#!/usr/bin/env python

###################################
# filename: dgstudent.py
# author: Leo Liberti
# date: 170302 (taken from dg.py)
# source: Python 2.7 (library)
# usage: put in same dir, import dgstudent, from dgstudent import *
###################################

############# SECT:IMPORTS #############
# basic
import sys         # system calls (.exit, .argv)
import re          # regular expression library, used in reading input files
import os.path
import time
import math        # standard math functions 
import numpy as np # the numeric library, with linear algebra calls
import scipy
from scipy import optimize # to use the optimization solvers in scipy 
from collections import OrderedDict # used to nice-print adjacency lists
import itertools   # find initial clique in DVOP order greedy alg
import operator    # operator.itemgetter() to find max in tuple components
import types
import warnings

# plots
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import proj3d
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches

# CPLEX
import cplex
from cplex.exceptions import CplexError

# for scipy optimizers
import functools     # automatic differentiation - Jacobian / Hessian (helper)
import algopy

# SDP
import cvxopt as cvx # the CvxOpt library, with various solvers including SDP
import picos as pic  # the modelling layer for CvxOpt

# options
np.set_printoptions(precision=2, suppress=True) # print precision for numpy

############# SECT:GLOBALS ###############

# set number of dimensions
global NumberOfDimensions
NumberOfDimensions = 3

# this is a technical hack to plot in 3D and update the plot after mouse ops
#   no need to change / update this; see
#   stackoverflow.com/questions/10374930/matplotlib-annotating-a-3d-scatter-plot
global labels_and_points
labels_and_points = []

# this is a tolerance for zero, used in next* computations
global myZero
myZero = 1e-6

# this is an upper bound on weights
global myInf
myInf = 1e+25

# maximum iteration for local NLP solvers 
nlpMaxIter = 100

# SDP solver tolerance (<0 to accept solver's default)
solverTolerance = 1.0e-3

############# SECT:CLASSES #############

## this class defines a vertex
##   a vertex has name, rank and an optional Euclidean position
class Vertex:
    # declare and initialize member attributes
    def __init__(self, nm):
        # a vertex is name and rank, and other optional fields
        self.name = nm
        self.rank = -1      # ranks start from 0
        self.pos = list()
    # printing vertices
    def __str__(self):
        return str(self.name)


## this class defines an edge or arc
##   an edge has a tail, a head, a lower/upper weights, a directed flag
##   optionally, also some chemical information for molecule graphs
##   for non-interval weights, wL represents the weight and wU is ignored
class Edge:
    # declare and initialize member attributes
    def __init__(self, v1, v2, w1, w2, givenDir):
        self.tail = v1
        self.head = v2
        self.wL = w1   
        self.wU = w2
        self.interval = 0
        if abs(w2 - w1) > myZero:
            self.interval = 1
        self.directed = givenDir
        # optional fields, only useful for .nmr files
        self.tlatom = ""
        self.hdatom = ""
        self.amino1 = ""
        self.amino2 = ""
    # printing edges
    def __str__(self):
        if self.directed == True:
            link = "->"
        else:
            link = "--"
        return str(self.tail) + link + str(self.head)
    # initialize the additional data structures in NMR files
    def nmr(self, atom1, atom2, am1, am2):
        self.tlatom = atom1
        self.hdatom = atom2
        self.amino1 = am1
        self.amino2 = am2

## this class defines a set of vertices
##   vertices are stored by name or by rank; this class provides the mappings
class Vertices:
    # initialize vertices from a set of edges
    def __init__(self, E):
        # create a set of vertex names
        self.names = set()
        # and fill it from the edges
        for e in E:
            self.names.add(e.tail.name)
            self.names.add(e.head.name)
        self.names = sorted(self.names, key=int)
        # create maps to find vertices from rank and name
        self.byRank = dict()
        self.byName = dict()
        i = 0   # this marks the fact that vertex ranks start from 0
        # fill these maps
        for vn in self.names:
            v = Vertex(vn)
            v.rank = i
            self.byRank[i] = v
            self.byName[vn] = v
            i = i + 1
        # set the correct vertex rank within the edge set
        for e in E:
            e.head.rank = self.byName[e.head.name].rank
            e.tail.rank = self.byName[e.tail.name].rank
        # cardinality of the vertex set
        self.card = i
    # print vertex sets
    def __str__(self):
        return " ".join(self.names)

## we don't define a set of edges, since it's simply a list of edges
##   it's the first data structure we read from input files
        
## this class defines a weighted graph
##   a graph has a Vertices object and a list of edges
##   it computes adjacency lists, which are somewhat cumbersome to use
##   example:
##     for u in range(n):
##       for v in G[G.V.byRank[u].name]:
##         # do something with (edge) G[G.V.byName[u].name][v]
class Graph:
    # declare and initialize member attributes
    def __init__(self, E):
        # vertices
        self.V = Vertices(E)
        # edges
        self.E = E
        # adjacency lists (implemented as a dict:V->(dicts:V->edge))
        #   maps v to dictionary mapping u to directed edge (v,u) and,
        #   if graph undirected, also store inverse edge in head list
        self.adj = dict()
        for v in self.V.names:
            self.adj[v] = dict()
        for e in E:
            self.adj[e.tail.name][e.head.name] = e
            if e.directed == 0:
                # inverse edge appearing in head list is directed as -1
                einv = Edge(e.head, e.tail, e.wL, e.wU, -1)
                einv.nmr(e.tlatom, e.hdatom, e.amino1, e.amino2)
                self.adj[e.head.name][e.tail.name] = einv
    # access the v-the adjacency list as G[v]
    def __getitem__(self, vname):
        return self.adj[vname]
    # return {u,v} if it is an edge, None otherwise
    def getEdge(self, u,v):
        try:
            e = self.adj[self.V.byRank[u].name].get(self.V.byRank[v].name)
        except KeyError, errmsg:
            e = None
        return e
    def getEdgeByName(self, u,v):
        return self.adj[u].get(v)
    def isEdge(self, u,v):
        try:
            e = self.adj[self.V.byRank[u].name].get(self.V.byRank[v].name)
        except KeyError, errmsg:
            e = None
        if (e != None):
            return True
        else:
            return False
    def isEdgeByName(self, u,v):
        e = self.adj[u].get(v)
        if (e != None):
            return True
        else:
            return False
    # return True iff {u,v} is an edge, where u,v are ranks
    def getEdgeWeight(self, u,v):
        try: 
            e = self.adj[self.V.byRank[u].name].get(self.V.byRank[v].name)
        except KeyError, errmsg:
            e = None
        if (e != None):
            return (e.wL, e.wU)
        else:
            return e
    # return the weight interval of the edge {u,v} where u,v are ranks
    # print the adjacency lists
    def __str__(self):
        adj = OrderedDict(sorted(self.adj.items(), key = lambda t : int(t[0])))
        ret = str()
        for v in adj:
            ret = ret + v + ":"
            for u in adj[v].values():
                s = " {0:s}({1:.2f})".format(u.head.name, round(u.wL, 2))
                ret = ret + s
            ret = ret + "\n";
        return ret
    # return a sparse version of the weighted adjacency matrix of the graph
    def sparseWeightedAdjMatrix(self):
        I = list()
        J = list()
        W = list()
        for e in self.E:
            i = e.tail.rank
            j = e.head.rank
            w = e.wL
            W.append(w)
            I.append(i)
            J.append(j)
            if e.directed == 0:
                W.append(w)
                I.append(j)
                J.append(i)
        return scipy.sparse.csr_matrix(scipy.sparse.coo_matrix((W,(I,J)), shape = (self.V.card,self.V.card)))
    # return the weighted adjacency matrix of the graph
    def weightedAdjMatrix(self):
        n = self.V.card
        A = myInf * np.ones((n,n))
        for i in range(n):
            A[i,i] = 0
        for e in self.E:
            A[e.tail.rank,e.head.rank] = e.wL
            A[e.head.rank,e.tail.rank] = e.wL
        return A
    # turn all wL=wU to (wL-err,wU+err) where err=percent*|wU|,
    #   return number of changed edges
    def intervalWeights(self, percent = 0.1):
        changed = 0
        for e in self.E:
            if abs(e.wU - e.wL) < myZero:
                err = percent*abs(e.wU)
                e.wL = e.wL - err
                e.wU = e.wU + err
                e.interval = 1
                changed = changed + 1
        return changed
    # plot the graph in 1D, 2D or 3D using the realization x
    def plot(self, x, edgeFlag=False):
        K = NumberOfDimensions
        fig = plt.figure()
        if K < 1:
            K = 1
        elif K > 3:
            K = 3
            x = x[:,0:K]
        if K == 1:
            # one-dimensional plots on a line y=1 in the xy plane
            ax = fig.add_subplot(111)
            # points and labels
            y = np.ones(len(x[:,0]))
            ax.plot(x[0,:], y, c='r', marker='o', linestyle='None')
            for label,lx,ly in zip(self.V.names,x[:,0], y):
                ax.annotate(label, xy=(lx,ly), xytext=(7,-3), textcoords='offset points')
            # edges as curved segments above and below
            if edgeFlag:
                i = 0
                for e in self.E:
                    updown = 1.5
                    if i % 2 == 1:
                        updown = 0.5
                    verts = [(x[e.tail.rank,0], 1.), (0.5*(x[e.tail.rank,0]+x[e.head.rank,0]), updown), (x[e.head.rank,0], 1.)]
                    codes = [Path.MOVETO, Path.CURVE3, Path.CURVE3]
                    path = Path(verts, codes)
                    patch = patches.PathPatch(path, facecolor='none', edgecolor='blue', lw=1)
                    ax.add_patch(patch)
                    i = i + 1
            # axes
            epsx = 0.1*(max(x[:,0])-min(x[:,0]))
            xmin = min(x[:,0]) - epsx
            xmax = max(x[:,0]) + epsx
            ax.axis([xmin, xmax, 0, 2])
        elif K == 2:
            # two-dimensional plots on the xy plane
            ax = fig.add_subplot(111)
            # points and labels
            ax.plot(x[:,0], x[:,1], c='r', marker='o', linestyle='None')
            for label,lx,ly in zip(self.V.names,x[:,0],x[:,1]):
                ax.annotate(label, xy=(lx,ly), xytext=(-6,6), textcoords='offset points')
            # edges as straight segments
            if edgeFlag:
                for e in self.E:
                    ax.plot([x[e.tail.rank,0], x[e.head.rank,0]], [x[e.tail.rank,1], x[e.head.rank,1]], c='b')
            # axes
            epsx = 0.1*(max(x[:,0])-min(x[:,0]))
            epsy = 0.1*(max(x[:,1])-min(x[:,1]))
            xmin = min(x[:,0]) - epsx
            xmax = max(x[:,0]) + epsx
            ymin = min(x[:,1]) - epsx
            ymax = max(x[:,1]) + epsx
            ax.axis([xmin, xmax, ymin, ymax])
        elif K == 3:
            # three-dimensional plots in xyz space
            ax = fig.add_subplot(111, projection='3d')
            # points and labels
            ax.scatter(x[:,0], x[:,1], x[:,2], c='r', marker='o')
            for label,lx,ly,lz in zip(self.V.names,x[:,0],x[:,1],x[:,2]):
                x2, y2, _ = proj3d.proj_transform(lx,ly,lz, ax.get_proj())
                lb = ax.annotate(label, xy=(x2,y2), xytext=(-6,6), textcoords='offset points')
                labels_and_points.append((lb, lx, ly, lz))
            # edges as straight segments
            if edgeFlag:
                for e in self.E:
                    ax.plot([x[e.tail.rank,0], x[e.head.rank,0]], [x[e.tail.rank,1], x[e.head.rank,1]], [x[e.tail.rank,2], x[e.head.rank,2]], c='b')
            # axes
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            def update_position(e):
                for lb, lx, ly, lz in labels_and_points:
                    x2, y2, _ = proj3d.proj_transform(lx, ly, lz, ax.get_proj())
                    lb.xy = x2,y2
                    lb.update_positions(fig.canvas.renderer)
                fig.canvas.draw()
            fig.canvas.mpl_connect('motion_notify_event', update_position) 
        plt.show()

## this class defines a DMDGP instance with a re-order
class DMDGP:
    # declare and initialize member attributes
    def __init__(self, G, R):
        self.reorder = R
        self.graph = G
        
## this class used to prettyprint list of floats
##   use as: print map(floatPrint, myList)
class floatPrint(float):
    def __repr__(self):
        return "%0.2f" % self
        
############# SECT:INSTANCE_INPUT ############

## read input file
##   this a dispatcher procedure
def edgesFromFile(inputFile):
    E = list()
    frm = fileFormat(inputFile)
    print "dgstudent:edgesFromFile(): input format is", frm
    if frm == "makegraph":
        E = readMakeGraphFile(inputFile)
    elif frm == "nmr":
        E = readNMRFile(inputFile)
    elif frm == "dimacs":
        E = readDIMACSFile(inputFile)
    elif frm == "txt":
        E = readTxtFile(inputFile)
    elif frm == "dat":
        E = readDatFile(inputFile)
    else:
        sys.exit("{0:s}: error: can't handle format \"{1:s}\" in input file {2:s}".format(sys.argv[0],frm,inputFile))
    return E

## read data file, decide input format
def fileFormat(filename):
    theFormat = "unknown"
    with open(filename) as f:
        for line in f:
            # look for first non-comment line
            line = line.strip()
            if len(line) > 0:
                if line[0] != '#':
                    # recognize format
                    if line == "Undirected;" or line == "Directed;":
                        # the .gph format (see makegraph.c)
                        theFormat = "makegraph"
                    elif line[0] == 'c' and line[1] == ' ':
                        # the well-known DIMACS format
                        theFormat = "dimacs"
                    elif line[0:5] == 'param':
                        # AMPL .dat file (automatically translated from gph)
                        theFormat = "dat"
                    elif line.lower() == "begin vertices":
                        # Andrea Cassioli's .dmdgp format
                        theFormat = "dmdgp"
                    else:
                        cols = line.split()
                        if len(cols) < 3:
                            theFormat = "unknown"
                        else:
                            m1 = re.search("[0-9]*", cols[0])
                            vtx1 = m1.group(0)
                            m2 = re.search("[0-9]*", cols[1])
                            vtx2 = m2.group(0)
                            m3 = re.search("[0-9]*\.?[0-9]*", cols[2])
                            ew1 = m3.group(0)
                            if len(cols) >= 4:
                                m4 = re.search("[0-9]*\.?[0-9]*", cols[3])
                                ew2 = m4.group(0)
                                if vtx1 == cols[0] and vtx2 == cols[1] and ew1 == cols[2] and ew2 == cols[3]:
                                    # the .nmr format (PDB)
                                    theFormat = "nmr"
                            elif vtx1 == cols[0] and vtx2 == cols[1] and ew1 == cols[2]:
                                # the simplest .txt format [vtx1 vtx2 weight12]
                                theFormat = "txt"
                            else:
                                theFormat = "unknown"
                    # we have the format, stop scanning
                    break
    return theFormat

## read graph (& reorder) from input file
def instanceFromFile(inputFile):
    frm = fileFormat(inputFile)
    R = list()
    if frm == "dmdgp":
        (E,R) = readDMDGPFile(inputFile)
    else:
        E = edgesFromFile(inputFile)
    G = Graph(E)
    n = G.V.card
    if len(R) == 0:
        # no given re-order, create the trivial one
        for u in range(n):
            R.append(G.V.byRank[u].name)
    return DMDGP(G,R)

## read the DIMACS graph file format
def readDIMACSFile(filename):
    dirFlag = False
    E = list()
    with open(filename) as f:
        for linecount, line in enumerate(f):
            line = line.strip()
            if len(line) > 0:
                # we're only interested in "arc" lines
                if line[0] == 'a':
                    cols = [c for c in line.split() if not '#' in c]
                    if len(cols) >= 5:
                        e = Edge(Vertex(cols[1]), Vertex(cols[2]), float(cols[3]), float(cols[4]), dirFlag)
                        if e.wL > e.wU:
                            print "readDIMACSFile: WARNING: interval weight[", e.wL, ",", e.wU, "] empty, setting to", e.wL
                            e.wU = e.wL
                    elif len(cols) >= 4:
                        e = Edge(Vertex(cols[1]), Vertex(cols[2]), float(cols[3]), float(cols[3]), dirFlag)
                    else:
                        print "readDIMACSFile: ERROR: arc line", linecount, "has < 4 columns"
                        exit('abort')
                    E.append(e)
    return E

## read data file written in .dmdgp format (format by Andrea Cassioli)
def readDMDGPFile(filename):
    dirFlag = False
    # we read F, then the BP order, then we permute edges and put them in E
    E = list()
    bporder = False
    order = list()
    section = "none"
    with open(filename) as f:
        for i,line in enumerate(f):
            line = line.strip()
            if len(line) > 0:
                if line[0] != '#':
                    # decide what section of the file we're in
                    if line.lower() == "begin vertices":
                        section = "vertices"
                    elif line.lower() == "end vertices":
                        section = "none"
                    elif line.lower() == "begin edges":
                        section = "edges"
                    elif line.lower() == "end edges":
                        section = "none"
                    elif line.lower() == "begin atom_names":
                        section = "atom_names"
                    elif line.lower() == "end atom_names":
                        section = "none"
                    elif line.lower() == "begin residues":
                        section = "residues"
                    elif line.lower() == "end residues":
                        section = "none"
                    elif line.lower() == "begin dihedral_angles":
                        section = "dihedral_angles"
                    elif line.lower() == "end dihedral_angles":
                        section = "none"
                    elif line.lower() == "begin bp_order":
                        section = "bp_order"
                    elif line.lower() == "end bp_order":
                        section = "none"
                    else:
                        # actions for each section
                        if section == "vertices":
                            # TO DO
                            a = False
                        elif section == "edges":
                            # read the (unpermuted) edge
                            cols = [c for c in line.split() if not '#' in c]
                            if len(cols) >= 5:
                                e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[3]), float(cols[4]), dirFlag)
                                if e.wL > e.wU:
                                    print "readDMDGPFile: WARNING: interval weight[", e.wL, ",", e.wU, "] empty, setting to", e.wL
                                    e.wU = e.wL
                            elif len(cols) >= 4:
                                e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[3]), float(cols[3]), dirFlag)
                            else:
                                print "readDMDGPFile: ERROR: line", linecount, "has < 4 columns"
                                exit('abort')                            
                            E.append(e)
                        elif section == "atom_names":
                            # TO DO
                            a = False
                        elif section == "residues":
                            # TO DO
                            a = False
                        elif section == "dihedral_angles":
                            # TO DO
                            a = False
                        elif section == "bp_order":
                            bporder = True
                            # vertices are stored as labels -- mostly, they
                            #   are integers, but they need not be
                            order.append(line)
    return (E,order)

## read data file written in AMPL .dat format
## (after line starting "param : E" there is a list of edges formatted as
##    i j w_{ij} I_{ij} )
def readDatFile(filename):
    dirFlag = False
    E = list()
    edgeflag = False
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0:                
                if line[0] != '#':
                    if edgeflag:
                        if line[0] == ';':
                            edgeflag = False
                        else:
                            cols = [c for c in line.split() if not '#' in c]
                            if len(cols) >= 6:
                                e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[4]), float(cols[5]), dirFlag)
                            elif len(cols) >= 4:
                                e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[2]), float(cols[2]), dirFlag)
                                if e.wL > e.wU:
                                    print "readDatFile: WARNING: interval weight[", e.wL, ",", e.wU, "] empty, setting to", e.wL
                                    e.wU = e.wL
                            else:
                                print "readDatFile: ERROR: line", linecount, "has < 4 columns"
                                exit('abort')
                            E.append(e)
                    else:
                        if line[0:9].replace(" ","") == 'param:E':
                            edgeflag = True
    return E

## read data file written by MakeGraph
def readMakeGraphFile(filename):
    dirFlag = False
    vtxFlag = True
    E = list()
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0:
                if line[0] != '#':
                    if vtxFlag == True:
                        # vertex part
                        if line == "Directed;":
                            dirFlag = True
                        elif line == "Edges:":
                            vtxFlag = False
                    else:
                        # edge part
                        cols = [c for c in line.split() if not '#' in c]
                        if len(cols) >= 4:
                            e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[2]), float(cols[3]), dirFlag)
                            E.append(e)
                        else:
                            print "readMakeGraphFile: ERROR: line", linecount, "has < 4 columns"
                            exit('abort')
    return E

## read data file written in .nmr format (PDB / Antonio Mucherino)
def readNMRFile(filename):
    dirFlag = False
    E = list()
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0:
                if line[0] != '#':
                    cols = [c for c in line.split() if not '#' in c]
                    e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[2]), float(cols[3]), dirFlag)
                    # look for optional columns
                    if len(cols) >= 8:
                        e.nmr(cols[4], cols[5], cols[6], cols[7])
                    elif len(cols) >= 7:
                        e.nmr(cols[4], cols[5], cols[6], "")
                    elif len(cols) >= 6:
                        e.nmr(cols[4], cols[5], "", "")
                    elif len(cols) >= 5:
                        e.nmr(cols[4], "", "", "")
                    elif len(cols) < 4:
                        print "readNMRFile: ERROR: line", linecount, "has < 4 columns"
                        exit('abort')
                    E.append(e)
    return E

## read data file written in .txt format (each line is "vtx1 vtx2 wL [wU]")
def readTxtFile(filename):
    dirFlag = False
    E = list()
    with open(filename) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0:
                if line[0] != '#':
                    cols = [c for c in line.split() if not '#' in c]
                    if len(cols) >= 4:
                        e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[2]), float(cols[3]), dirFlag)
                    elif len(cols) >= 3:
                        e = Edge(Vertex(cols[0]), Vertex(cols[1]), float(cols[2]), float(cols[2]), dirFlag)
                    else:
                        print "readTxtFile: ERROR: line", linecount, "has < 3 columns"
                        exit('abort')
                    E.append(e)
    return E

## read realization from _rlz.dat file 
def realizationFromRlzDatFile(G, inputFile):
    n = G.V.card
    K = NumberOfDimensions
    x = np.zeros((n,K))
    f = open(inputFile, "r")
    offset = 1 # start counting vertices from 1 by default
    idx = 0;
    for line in f:
        v = line.split()
        # assumption: if line has exactly K+1 components, it's a point
        if len(v) == K+1:
            for k in range(1,K+1):
                x[idx,k-1] = float(v[k])
            idx = idx + 1
    return x

############# SECT:BASIC_METHODS ############

## compute the mean scaled distance error of x w.r.t. weights in G
##    p  is the order of the norm {1,2,inf}
def MDE(G,x,p = 2, scaled=0):
    mde = 0
    for e in G.E:
        dij = np.linalg.norm(np.subtract(x[e.tail.rank,:], x[e.head.rank,:]), ord=p)
        if e.interval == 0:
            newmde = abs(dij - e.wL)
            if scaled == 1:
                scaling = e.wL
                if scaling < 1:
                    scaling = 1
                newmde = newmde / scaling
        else:
            newmde = max(0, e.wL - dij) + max(0, dij - e.wU)
            if scaled == 1:
                scalingL = e.wL
                if scL < 1:
                    scL = 1
                scalingU = e.wU
                if scU < 1:
                    scU = 1
                newmde = max(0, e.wL - dij) / scL + max(0, dij - e.wU) / scU
        mde = mde + newmde
    return mde / len(G.E)

## compute the largest scaled distance error of the w.r.t. weights in G
##    p  is the order of the norm {1,2,inf}
def LDE(G,x,p = 2, scaled = 0):
    lde = 0
    for e in G.E:
        dij = np.linalg.norm(np.subtract(x[e.tail.rank,:], x[e.head.rank,:]), ord=p)
        if e.interval == 0:
            newlde = abs(dij - e.wL)
            if scaled == 1:
                scaling = e.wL
                if scaling < 1:
                    scaling = 1
                newlde = newlde / scaling
        else:
            newlde = max(0, e.wL - dij) + max(0, dij - e.wU)
            if scaled == 1:
                scalingL = e.wL
                if scL < 1:
                    scL = 1
                scalingU = e.wU
                if scU < 1:
                    scU = 1
                newlde = max(0, e.wL - dij) / scL + max(0, dij - e.wU) / scU
                #print "{0:s}\t{1:s}\t{2:.2f}\t{3:.2f}\t{4:d}\t{5:d}".format(e.tail.name, e.head.name, newlde, e.wL*e.wL, e.tail.rank, e.head.rank)
        if lde < newlde:
            lde = newlde
            #print "edge", e.tail.rank, "->", e.head.rank, "new lde", newlde, ":", e.wL, dij, e.wU
    return lde 

## return a random realization
def RandomRealization(G, K, M):
    return np.random.uniform(-M,M,(K,G.V.card))

## return the distance matrix of a realization
##   p is the norm order
def distanceMatrix(x, p=2):
    n = len(x[:,0])
    EDM = np.zeros((n,n))
    for u in range(n-1):
        for v in range(u+1,n):
            EDM[u,v] = np.linalg.norm(np.subtract(x[u,:],x[v,:]), ord=p)
            EDM[v,u] = EDM[u,v]
    return EDM

## is given matrix PSD?
def isPSD(A):
  eigvals, eigvects = np.linalg.eigh(A)
  return np.all(eigvals > -myZero)

## get the maximum edge weight from G
def maxEdgeWeight(G):
    return partialDistanceMatrix(G).max()

## count near-zeros of a matrix
def numNearZeros(A):
    ret = 0
    (m,n) = A.shape
    for i in range(m):
        for j in range(n):
            if abs(A[i,j]) <= myZero:
                ret = ret + 1
    return ret

## form the partial distance matrix from a graph
##   missing non-diagonal entries have a zero 
def partialDistanceMatrix(G):
    n = G.V.card
    pEDM = np.zeros((n,n))
    for u in range(n):
        for v in G[G.V.byRank[u].name]:
            pEDM[u,G.V.byName[v].rank] = G[G.V.byRank[u].name][v].wL
    return pEDM

## (random) r-diagonal matrices, see [Barvinok DCG 1995, example 4.1]
def rDiagonal(n,r):
    F = np.identity(n)
    for i in range(n):
        for j in range(r):
            if i+j+1 < n:
                rndeps = np.random.uniform(-1.0/r,1.0/r)
                F[i,i+j+1] = rndeps
                F[i+j+1,i] = rndeps
    return F

## random square positive definite matrix
def rndPositiveDefinite(n):
    F = np.identity(n)
    for i in range(n):
        F[i,i] = np.random.uniform(1.0,5.0)
    P = np.zeros((n,n))
    for i in range(n):
        for j in range(i+1,n):
            P[i,j] = np.random.uniform(1.0,5.0)
    return np.dot(P.T,np.dot(F,P))

## return the squared distance matrix of a realization
def squareDistanceMatrix(x):
    return np.square(distanceMatrix(x))

## get the sum of all the edge weights (in abs val) of G
def sumEdgeWeight(G):
    return sum(e.wL for e in G.E)

## Floyd-Warshall:
##   return EDM with missing edge weights replaced by shortest path lengths 
def FloydWarshall(G):
    n = G.V.card
    M = maxEdgeWeight(G)
    EDM = (M+1)*np.ones((n,n))
    # initialize with existing weights
    for u in range(n):
        # zero diagonal
        EDM[u,u] = 0
        for v in G[G.V.byRank[u].name]:
            EDM[u,G.V.byName[v].rank] = G[G.V.byRank[u].name][v].wL
    for u in range(n):
        for v in range(n):
            for w in range(n):
                d = EDM[u,w] + EDM[w,v]
                if EDM[u,v] > d:
                    EDM[u,v] = d
    return EDM

## Floyd-Warshall:
##   return EDM with missing edge weights replaced by shortest path lengths
##   this version has a default big M = (sum edge weights)
def FloydWarshall1(G):
    n = G.V.card
    M = sumEdgeWeight(G)
    EDM = (M+1)*np.ones((n,n))
    # initialize with existing weights
    for u in range(n):
        # zero diagonal
        EDM[u,u] = 0
        for v in G[G.V.byRank[u].name]:
            EDM[u,G.V.byName[v].rank] = G[G.V.byRank[u].name][v].wL
    for u in range(n):
        for v in range(n):
            for w in range(n):
                d = EDM[u,w] + EDM[w,v]
                if EDM[u,v] > d:
                    EDM[u,v] = d
    return EDM

## a very simply multistart metaheuristic based on IPOpt
def MultiStart(G, iterations, method="ipopt"):
    # size
    K = NumberOfDimensions
    n = G.V.card
    # bounds [-M,M]
    A = FloydWarshall(G)
    M = max(A.flatten())
    # obj fun
    ldestar = sum(A.flatten())    
    # MS
    print "MultiStart(", method, "):"
    print "itn ldeinit lde ldestar"
    for t in range(iterations):
        xinit = RandomRealization(G,K,M)
        ldeinit = LDE(G,xinit)
        x = LocalNLPSolver(G,xinit,K,{},method)
        lde = LDE(G,x)
        if lde < ldestar:
            ldestar = lde
            xstar = x
        print t, ldeinit, lde, ldestar
    return xstar

## factor a square matrix 
def factor(A):
    n = A.shape[0]
    (evals,evecs) = np.linalg.eigh(A)
    evals[evals < 0] = 0  # closest SDP matrix
    X = evecs #np.transpose(evecs)
    sqrootdiag = np.eye(n)
    for i in range(n):
        sqrootdiag[i,i] = math.sqrt(evals[i])
    X = X.dot(sqrootdiag)
    return np.fliplr(X)

## spanning tree based constructive heuristic
##   grow a breadth-first spanning tree based on largest degree first
##   realize vertices in the BFS order of the tree, each vertex randomly
##   on the sphere S(parent, edge length)
def SpanningTreeEmbedding(G,K):
    n = G.V.card
    K = NumberOfDimensions
    x = np.zeros((n,K))
    marked = np.zeros(n)  # marked vertices are already in the tree
    degSeq = [(len(G[G.V.byRank[u].name]),u) for u in range(n)]
    maxDeg = max(degSeq)
    deg = dict()
    for (d,v) in degSeq:
        deg[v] = d
    q = Q.PriorityQueue()
    q.put((-maxDeg[0],maxDeg[1]))
    R = set()
    R.add(maxDeg[1])
    while not q.empty():
        v = q.get()[1]
        for wname in G[G.V.byRank[v].name]:
            w = G.V.byName[wname].rank
            if w not in R:
                if G.isEdge(v,w):
                    dist = G.getEdgeWeight(v,w)
                elif G.isEdge(w,v):
                    dist = G.getEdgeWeight(w,v)
                rnddir = np.random.uniform(-1,1,K)
                rnddir /= np.linalg.norm(rnddir)
                rnddir *= 0.5*(dist[0]+dist[1])
                x[w,:] = np.add(x[v,:],rnddir)
                R.add(w)
                q.put((-(deg[w]),w))
    return x

############# SECT:FORMULATION_BASED #############


## call to local solver for min sum_uv (||xu-xv||^2 - d_uv^2)^2
##   G is the input graph, xinit is the "starting point", and
##   xfixed is a map vertex->points to be fixed before optimization
##     this allows optimizing only some points in the realization
##
def LocalNLPSolver(G, xinit, K2=NumberOfDimensions, xfixed={}, method="lbfgs", maxitn=nlpMaxIter):
    # this class serves the purpose of making member attributes available
    #    to subfunctions (just a tech detail to overcome a Python 2.x limit)
    class local:
        gph = G
        fxd = xfixed # this is a dictionary of lists not a list of lists
                     # initialize to [] if no fixed vertices
        K = K2
        
    # evaluate objective function sum_uv (||x_u-x_v||^2-d_uv^2)^2
    #    warning: x is a flat array, needs to be mapped to a structured
    #       array with indices in V and K
    def eval_fun(x):
        for v in local.fxd:
            for k in range(local.K):
                x[v*local.K + k] = local.fxd[v][k]
        ret = 0
        for e in local.gph.E:
            dij2 = sum((x[e.head.rank*local.K+l] - x[e.tail.rank*local.K+l])**2 for l in range(local.K))
            dij = np.sqrt(dij2)
            if e.interval == 0:
                ret = ret + (dij2 - e.wL**2)**2
            else:
                ret = ret + (max(0,e.wL**2-dij2))**2 + (max(0,dij2-e.wU**2))**2
        return ret

    # return |E|-vector ||x_u-x_v||^2-d_uv^2
    #    warning: x is a flat array, needs to be mapped to a structured
    #       array with indices in V and K
    def eval_sys(x):
        for v in local.fxd:
            for k in range(local.K):
                x[v*local.K + k] = local.fxd[v][k]
        ret = list()
        for e in local.gph.E:
            dij2 = sum((x[e.head.rank*local.K+l] - x[e.tail.rank*local.K+l])**2 for l in range(local.K))
            dij = np.sqrt(dij2)
            if e.interval == 0:
                ret.append(dij - e.wL)
            else:
                ret.append(max(0,e.wL - dij) + max(0,dij - e.wU))
        return np.array(ret)
    
    # actually do the work
    xflatinit = xinit.flatten()
    print "dgstudent:LocalNLPSolver: calling", method, "for", maxitn, "iterations"

    if method == "lbfgs":
        sol = optimize.minimize(eval_fun, xflatinit, jac=functools.partial(eval_grad, eval_fun), method='L-BFGS-B', options={'disp':True, 'maxiter': maxitn})
        X = sol.x        
    elif method == "cg":
        sol = optimize.minimize(eval_fun, xflatinit, jac=functools.partial(eval_grad, eval_fun), method='CG', options={'disp':True, 'maxiter':maxitn})
        X = sol.x
    elif method == "neldermead":
        sol = optimize.minimize(eval_fun, xflatinit, method='Nelder-Mead', options={'disp': True, 'maxiter':maxitn})
        X = sol.x
    elif method == "leastsq":
        sol = optimize.leastsq(eval_sys, xflatinit, ftol = 1e-2, xtol = 1e-2, maxfev = maxitn)
        X = sol[0]
    elif method == "root":
        sol = optimize.root(eval_sys, xflatinit, method='lm', options={'ftol':1e-2, 'xtol':1e-3, 'maxiter':maxitn})
        X = sol.x
    else:
        print "dgstudent:LocalNLPSolver: method", method, " not known"
        sys.exit(115)
    xstar = np.reshape(X, (local.gph.V.card, local.K), order='C')
    return xstar
# used to compute the gradient
def eval_grad(f, theta):
    theta = algopy.UTPM.init_jacobian(theta)
    return algopy.UTPM.extract_jacobian(f(theta))
# used to compute the Hessian
def eval_hess(f, theta):
    theta = algopy.UTPM.init_hessian(theta)
    return algopy.UTPM.extract_hessian(len(theta), f(theta))    


